/*
 * arduinofreq
 * Callibrate temperature dependence of Arduino on-board clock frequency.
 *
 * Connections:
 *   digital input 2  = serial line from GPS receiver (4800 bps)
 *   digital output 3 = serial line to GPS receiver (idle high)
 *   digital output 4 = pull low to enable GPS receiver
 *   digital input 8  = GPS pulse per second
 *   analog pin 0     = random digital output 1 bps
 *
 * Serial output (115200 bps):
 *
 *   Pulse captured (8 bytes):
 *     byte 0      = 0x05
 *     bytes 1..4  = (uint32_t) time stamp
 *     bytes 5..6  = (uint16_t) temperature sensor (sum of 4 samples)
 *     byte 7      = (uint8_t)  state of digital output pin
 *
 *   Forward output from GPS receiver (2 bytes):
 *     byte 0      = 0x04
 *     byte 1      = byte received from GPS
 *
 *   Multi-byte values are transmitted with LSB first.
 *
 * The sketch generates a random digital output signal on analog pin 0
 * at 1 bit per second. If this pin is connected via a resistor to ground,
 * it may be used to test stability of the oscillator when variable
 * currents are flowing on the board.
 */

#include <avr/io.h>

const long gpsrx_baud = 4800;
volatile uint8_t gpsrx_bitcnt = 0;
volatile uint8_t gpsrx_byte = 0;
volatile uint8_t gpsrx_buf = 0;
volatile uint8_t gpsrx_available = 0;

volatile uint16_t timecnt = 0;          // bits 16..31 of global time counter

volatile uint8_t pulse_pending = 0;     // 1 if a captured pulse is pending
volatile uint32_t pulse_time = 0;       // time of last captured pulse

uint8_t varload = 0;


// Handle level change on GPS RX pin.
ISR(PCINT2_vect) {

  // detect beginning of start bit if port is idle and falling edge seen
  if (gpsrx_bitcnt == 0) {

    // read input level
    uint8_t c = (PIND & 4);

    if (c == 0) {
      gpsrx_bitcnt = 1;

      // disable further pin-change interrupts
      PCMSK2 = 0;
        
      // set timer to capture first data bit
      TCNT2  = 250000 / gpsrx_baud + 2;  // count half period to skip start bit
      TIFR2  = bit(OCF2A);               // clear pending interrupt
      TIMSK2 = bit(OCIE2A);              // enable interrupt on counter match
    }
  }

}


// Timer2 counter match interrupt
ISR(TIMER2_COMPA_vect) {

  // read input level
  uint8_t c = (PIND & 4);

  if (gpsrx_bitcnt == 10) {

    // receive stop bit
    if (c != 0) {
      // got good character
      gpsrx_buf = gpsrx_byte;
      gpsrx_available = 1;
    }

    // disable timer interrupt; stay idle until next byte
    gpsrx_bitcnt = 0;
    
    // disable timer interrupt; enable pin-change interrupt
    TIMSK2 = 0;    
    PCMSK2 = bit(PCINT18);

  } else if (gpsrx_bitcnt != 0) {
    
    // receive next data bit
    if (c)
      gpsrx_byte = (gpsrx_byte >> 1) | 0x80;
    else
      gpsrx_byte >>= 1;

    gpsrx_bitcnt++;
  }
  
}


// Timer1 input capture interrupt
ISR(TIMER1_CAPT_vect) {

  // get time of captured event
  uint8_t  t0 = ICR1L;
  uint8_t  t1 = ICR1H;
  uint16_t tcnt = timecnt;

  // compensate for pending overflow interrupt
  if (t1 < 0x40 && (TIFR1 & bit(TOV1)))
    tcnt++;
  
  // store time of captured event
  pulse_time = ((uint32_t)tcnt << 16) | ((uint32_t)t1 << 8) | t0;

  // set flag to indicate pulse captured
  pulse_pending = 1;
}


// Timer1 overflow interrupt
ISR(TIMER1_OVF_vect) {
  // increment time counter
  timecnt++;
}


// Read internal AVR temperature sensor.
static uint16_t get_temp_sensor() {
  
  // setup ADC
  ADMUX  = bit(REFS1)|bit(REFS0)|   // select internal 1.1V reference
           bit(MUX3);               // select input channel 8 = temp sensor
  ADCSRA = bit(ADEN)|               // enable ADC, no auto-trigger, no interrupt
           bit(ADPS2)|bit(ADPS1)|bit(ADPS0); // select prescaling factor 128 (125 kHz)

  // start conversion
  ADCSRA |= bit(ADSC);

  // wait until conversion finished
  while (bit_is_set(ADCSRA, ADSC)) ;

  // read value and return it; MUST read ADCL before ADCH
  uint8_t low, high;
  low = ADCL;
  high = ADCH;

  return (high << 8) | low;
}


// write a 16-bit integer to the serial port (LSB first)
static void serial_write_int16(uint16_t v)
{
    Serial.write(v & 0xff);
    Serial.write((v >> 8) & 0xff);
}


// write a 32-bit integer to the serial port (LSB first)
static void serial_write_int32(uint32_t v)
{
    Serial.write(v & 0xff);
    Serial.write((v >> 8) & 0xff);
    Serial.write((v >> 16) & 0xff);
    Serial.write((v >> 24) & 0xff);
}


void setup() {

  // serial port to 115200 bps
  Serial.begin(115200);

  // set pin 2 to INPUT for GPS RX receiving; enable pullup.
  pinMode(2, INPUT);
  digitalWrite(2, HIGH);

  // set pin 3 for OUTPUT to GPS receiver (idle serial line)
  // NOTE: This is absolutely essential. Without high signal
  // on this pin, the receiver does not output correct NMEA strings.
  pinMode(3, OUTPUT);
  digitalWrite(3, HIGH);

  // pull pin 4 low to enable the GPS receiver
  pinMode(4, OUTPUT);
  digitalWrite(4, LOW);

  // prepare pin A0 for random digital output
  pinMode(A0, OUTPUT);
  digitalWrite(A0, LOW);

  // prepare timer2 for GPS RX receiving
  TIMSK2 = 0;                        // disable interrupts
  TCCR2A = bit(WGM21);               // CTC mode
  TCCR2B = bit(CS21) | bit(CS20);    // use system clock/32
  OCR2A  = 500000 / gpsrx_baud - 1;  // interrupt interval equal to 1 bit period

  // enable interrupt on digital input pin 2 (GPS RX)
  PCICR  |= bit(PCIE2);   // enable PCINT23..16
  PCMSK2 |= bit(PCINT18); // enable PCINT18 = digital input 2

  // disable analog comparator, use ICFP1 for input capturing
  ACSR  = bit(ACD);
  
  // setup timer1 for time keeping and input capturing
  TCCR1A = 0;                     // normal mode, count to 0xffff
  TCCR1B = bit(ICNC1)|bit(ICES1)|bit(CS10); // no prescaling, input capture on rising edge
  TIMSK1 = bit(ICIE1)|bit(TOIE1); // enable interrupt on input capture and on overflow
}


void loop() {

  // write pending events to serial port
  if (pulse_pending) {
    // write time stamp and AVR temperature to serial port
    Serial.write(0x05);
    serial_write_int32(pulse_time);
    uint16_t temp = 0;
    temp += get_temp_sensor();
    temp += get_temp_sensor();
    temp += get_temp_sensor();
    temp += get_temp_sensor();
    serial_write_int16(temp);      
    Serial.write(varload);
    pulse_pending = 0;

    // write random level to output pin A0
    varload = random(2);
    digitalWrite(A0, varload);
  }

  // forward data from GPS receiver
  if (gpsrx_available) {
    uint8_t c = gpsrx_buf;
    gpsrx_available = 0;
    Serial.write(0x04);
    Serial.write(c);
  }

}

